import numpy
from numpy import array, zeros
from interpolation.smolyak import SmolyakGrid as SmolyakGrid0
from interpolation.smolyak import SmolyakInterp, build_B
from dolo.numeric.grids import cat_grids, n_nodes, node
from dolo.numeric.grids import UnstructuredGrid, CartesianGrid, SmolyakGrid, EmptyGrid
from dolo.numeric.misc import mlinspace
import scipy
from dolo.numeric.grids import *

# from dolo.numeric.decision_rule import CallableDecisionRule, cat_grids
import numpy as np

import numpy as np


def filter_controls(a, b, ndims, controls):

    from interpolation.splines.filter_cubic import filter_data, filter_mcoeffs

    dinv = (b - a) / (ndims - 1)
    ndims = array(ndims)
    n_m, N, n_x = controls.shape
    coefs = zeros((n_m,) + tuple(ndims + 2) + (n_x,))
    for i_m in range(n_m):
        tt = filter_mcoeffs(a, b, ndims, controls[i_m, ...])
        # for i_x in range(n_x):
        coefs[i_m, ...] = tt
    return coefs


class Linear:
    pass


class Cubic:
    pass


class Chebychev:
    pass


interp_methods = {
    "cubic": Cubic(),
    "linear": Linear(),
    "multilinear": Linear(),
    "chebychev": Chebychev(),
}

###


class CallableDecisionRule:
    def __call__(self, *args):
        args = [np.array(e) for e in args]
        if len(args) == 1:
            if args[0].ndim == 1:
                return self.eval_s(args[0][None, :])[0, :]
            else:
                return self.eval_s(args[0])
        elif len(args) == 2:
            if args[0].dtype in ("int64", "int32"):
                (i, s) = args
                if s.ndim == 1:
                    return self.eval_is(i, s[None, :])[0, :]
                else:
                    return self.eval_is(i, s)
                return self.eval_is()
            else:
                (m, s) = args[0], args[1]
                if s.ndim == 1 and m.ndim == 1:
                    return self.eval_ms(m[None, :], s[None, :])[0, :]
                elif m.ndim == 1:
                    m = m[None, :]
                elif s.ndim == 1:
                    s = s[None, :]
                return self.eval_ms(m, s)


class DecisionRule(CallableDecisionRule):

    exo_grid: Grid
    endo_grid: Grid

    def __init__(
        self,
        exo_grid: Grid,
        endo_grid: Grid,
        interp_method="cubic",
        dprocess=None,
        values=None,
    ):

        if interp_method not in interp_methods.keys():
            raise Exception(
                f"Unknown interpolation type: {interp_method}. Try one of: {tuple(interp_methods.keys())}"
            )

        self.exo_grid = exo_grid
        self.endo_grid = endo_grid
        self.interp_method = interp_method
        self.dprocess = dprocess

        self.__interp_method__ = interp_methods[interp_method]

        # here we could replace with a caching mechanism resolving dispatch in advance
        self.__eval_ms__ = eval_ms
        self.__eval_is__ = eval_is
        self.__eval_s__ = eval_s
        self.__get_coefficients__ = get_coefficients

        if values is not None:
            self.set_values(values)

    def set_values(self, x):
        self.coefficients = self.__get_coefficients__(
            self, self.exo_grid, self.endo_grid, self.__interp_method__, x
        )

    def eval_ms(self, m, s):
        return self.__eval_ms__(
            self, self.exo_grid, self.endo_grid, self.__interp_method__, m, s
        )

    def eval_is(self, i, s):
        return self.__eval_is__(
            self, self.exo_grid, self.endo_grid, self.__interp_method__, i, s
        )

    def eval_s(self, s):
        return self.__eval_s__(
            self, self.exo_grid, self.endo_grid, self.__interp_method__, s
        )

    def eval_ijs(self, i, j, s):

        if isinstance(self.exo_grid, UnstructuredGrid):
            out = self.eval_is(j, s)
        elif isinstance(self.exo_grid, EmptyGrid):
            out = self.eval_s(s)
        elif isinstance(self.exo_grid, CartesianGrid):
            m = self.dprocess.inode(i, j)[None, :].repeat(s.shape[0], axis=0)
            out = self.eval_ms(m, s)
        else:
            raise Exception("Not Implemented.")

        return out


# this is *not* meant to be used by users

from multipledispatch import dispatch

namespace = dict()
multimethod = dispatch(namespace=namespace)

# Cartesian x Cartesian x Linear


@multimethod
def get_coefficients(
    itp: object,
    exo_grid: CartesianGrid,
    endo_grid: CartesianGrid,
    interp_type: Linear,
    x: object,
):
    grid = exo_grid + endo_grid
    xx = x.reshape(tuple(grid.n) + (-1,))
    return xx.copy()


@multimethod
def eval_ms(
    itp: object,
    exo_grid: CartesianGrid,
    endo_grid: CartesianGrid,
    interp_type: Linear,
    m: object,
    s: object,
):

    assert m.ndim == s.ndim == 2

    grid = exo_grid + endo_grid  # one single CartesianGrid

    coeffs = itp.coefficients

    gg = grid.__numba_repr__()
    from interpolation.splines import eval_linear

    x = np.concatenate([m, s], axis=1)

    return eval_linear(gg, coeffs, x)


@multimethod
def eval_is(
    itp: object,
    exo_grid: CartesianGrid,
    endo_grid: CartesianGrid,
    interp_type: Linear,
    i: object,
    s: object,
):
    m = exo_grid.node(i)[None, :]
    return eval_ms(itp, exo_grid, endo_grid, interp_type, m, s)


# Cartesian x Cartesian x Cubic


@multimethod
def get_coefficients(
    itp: object,
    exo_grid: CartesianGrid,
    endo_grid: CartesianGrid,
    interp_type: Cubic,
    x: object,
):

    from interpolation.splines.prefilter_cubic import prefilter_cubic

    grid = exo_grid + endo_grid  # one single CartesianGrid
    x = x.reshape(tuple(grid.n) + (-1,))
    gg = grid.__numba_repr__()
    return prefilter_cubic(gg, x)


@multimethod
def eval_ms(
    itp: object,
    exo_grid: CartesianGrid,
    endo_grid: CartesianGrid,
    interp_type: Cubic,
    m: object,
    s: object,
):

    from interpolation.splines import eval_cubic

    assert m.ndim == s.ndim == 2

    grid = exo_grid + endo_grid  # one single CartesianGrid
    coeffs = itp.coefficients

    gg = grid.__numba_repr__()

    x = np.concatenate([m, s], axis=1)

    return eval_cubic(gg, coeffs, x)


@multimethod
def eval_is(
    itp: object,
    exo_grid: CartesianGrid,
    endo_grid: CartesianGrid,
    interp_type: Cubic,
    i: object,
    s: object,
):
    m = exo_grid.node(i)[None, :]
    return eval_ms(itp, exo_grid, endo_grid, interp_type, m, s)


# UnstructuredGrid x Cartesian x Linear


@multimethod
def get_coefficients(
    itp: object,
    exo_grid: UnstructuredGrid,
    endo_grid: CartesianGrid,
    interp_type: Linear,
    x: object,
):
    return [x[i].reshape(tuple(endo_grid.n) + (-1,)).copy() for i in range(x.shape[0])]


@multimethod
def eval_is(
    itp: object,
    exo_grid: UnstructuredGrid,
    endo_grid: CartesianGrid,
    interp_type: Linear,
    i: object,
    s: object,
):

    from interpolation.splines import eval_linear

    assert s.ndim == 2
    coeffs = itp.coefficients[i]
    gg = endo_grid.__numba_repr__()

    return eval_linear(gg, coeffs, s)


# UnstructuredGrid x Cartesian x Cubic


@multimethod
def get_coefficients(
    itp: object,
    exo_grid: UnstructuredGrid,
    endo_grid: CartesianGrid,
    interp_type: Cubic,
    x: object,
):
    from interpolation.splines.prefilter_cubic import prefilter_cubic

    gg = endo_grid.__numba_repr__()
    return [
        prefilter_cubic(gg, x[i].reshape(tuple(endo_grid.n) + (-1,)))
        for i in range(x.shape[0])
    ]


@multimethod
def eval_is(
    itp: object,
    exo_grid: UnstructuredGrid,
    endo_grid: CartesianGrid,
    interp_type: Cubic,
    i: object,
    s: object,
):

    from interpolation.splines import eval_cubic

    assert s.ndim == 2
    coeffs = itp.coefficients[i]
    gg = endo_grid.__numba_repr__()
    return eval_cubic(gg, coeffs, s)


# UnstructuredGrid x Cartesian x Linear


@multimethod
def get_coefficients(
    itp: object,
    exo_grid: UnstructuredGrid,
    endo_grid: CartesianGrid,
    interp_type: Linear,
    x: object,
):
    return [x[i].copy() for i in range(x.shape[0])]


@multimethod
def eval_is(
    itp: object,
    exo_grid: UnstructuredGrid,
    endo_grid: CartesianGrid,
    interp_type: Linear,
    i: object,
    s: object,
):

    from interpolation.splines import eval_linear

    assert s.ndim == 2

    coeffs = itp.coefficients[i]
    gg = endo_grid.__numba_repr__()

    return eval_linear(gg, coeffs, s)


# Empty x Cartesian x Linear


@multimethod
def get_coefficients(
    itp: object,
    exo_grid: EmptyGrid,
    endo_grid: CartesianGrid,
    interp_type: Linear,
    x: object,
):
    grid = exo_grid + endo_grid
    xx = x.reshape(tuple(grid.n) + (-1,))
    return xx.copy()


@multimethod
def eval_s(
    itp: object,
    exo_grid: EmptyGrid,
    endo_grid: CartesianGrid,
    interp_type: Linear,
    s: object,
):
    from interpolation.splines import eval_linear

    assert s.ndim == 2
    coeffs = itp.coefficients
    gg = endo_grid.__numba_repr__()
    return eval_linear(gg, coeffs, s)


# Empty x Cartesian x Cubic


@multimethod
def get_coefficients(
    itp: object,
    exo_grid: EmptyGrid,
    endo_grid: CartesianGrid,
    interp_type: Cubic,
    x: object,
):
    from interpolation.splines.prefilter_cubic import prefilter_cubic

    grid = endo_grid  # one single CartesianGrid
    gg = endo_grid.__numba_repr__()
    return prefilter_cubic(gg, x[0].reshape(tuple(grid.n) + (-1,)))


@multimethod
def eval_s(
    itp: object,
    exo_grid: EmptyGrid,
    endo_grid: CartesianGrid,
    interp_type: Cubic,
    s: object,
):
    from interpolation.splines import eval_cubic

    assert s.ndim == 2
    coeffs = itp.coefficients
    gg = endo_grid.__numba_repr__()
    return eval_cubic(gg, coeffs, s)


## an empty grid can be indexed by an integer or a vector


@multimethod
def eval_is(
    itp: object,
    exo_grid: EmptyGrid,
    endo_grid: object,
    interp_type: object,
    i: object,
    s: object,
):
    return eval_s(itp, exo_grid, endo_grid, interp_type, s)


@multimethod
def eval_ms(
    itp: object,
    exo_grid: EmptyGrid,
    endo_grid: object,
    interp_type: object,
    m: object,
    s: object,
):
    return eval_s(itp, exo_grid, endo_grid, interp_type, s)


####


class ConstantDecisionRule(CallableDecisionRule):
    def __init__(self, x0):
        self.x0 = x0

    def eval_s(self, s):
        if s.ndim == 1:
            return self.x0
        else:
            N = s.shape[0]
            return self.x0[None, :].repeat(N, axis=0)

    def eval_is(self, i, s):
        return self.eval_s(s)

    def eval_ms(self, m, s):
        return self.eval_s(s)


import dolang
import dolang.symbolic
from dolang.symbolic import stringify_symbol
from dolo.numeric.decision_rule import CallableDecisionRule
from dolang.factory import FlatFunctionFactory
from dolang.function_compiler import make_method_from_factory


class CustomDR(CallableDecisionRule):
    def __init__(self, values, model=None):

        from dolang.symbolic import sanitize, stringify

        exogenous = model.symbols["exogenous"]
        states = model.symbols["states"]
        controls = model.symbols["controls"]
        parameters = model.symbols["parameters"]

        preamble = dict([(s, values[s]) for s in values.keys() if s not in controls])
        equations = [values[s] for s in controls]

        variables = exogenous + states + controls + [*preamble.keys()]

        preamble_str = dict()

        for k in [*preamble.keys()]:
            v = preamble[k]
            if "(" not in k:
                vv = f"{k}(0)"
            else:
                vv = k

            preamble_str[stringify(vv)] = stringify(sanitize(v, variables))

        # let's reorder the preamble
        from dolang.triangular_solver import get_incidence, triangular_solver

        incidence = get_incidence(preamble_str)
        sol = triangular_solver(incidence)
        kk = [*preamble_str.keys()]
        preamble_str = dict([(kk[k], preamble_str[kk[k]]) for k in sol])

        equations = [
            dolang.symbolic.sanitize(eq, variables=variables) for eq in equations
        ]
        equations_strings = [dolang.stringify(eq) for eq in equations]

        args = dict(
            [
                ("m", [(e, 0) for e in exogenous]),
                ("s", [(e, 0) for e in states]),
                ("p", [e for e in parameters]),
            ]
        )

        args = dict([(k, [stringify_symbol(e) for e in v]) for k, v in args.items()])

        targets = [stringify_symbol((e, 0)) for e in controls]

        eqs = dict([(targets[i], eq) for i, eq in enumerate(equations_strings)])

        fff = FlatFunctionFactory(preamble_str, eqs, args, "custom_dr")

        fun, gufun = make_method_from_factory(fff)

        self.p = model.calibration["parameters"]
        self.exo_grid = model.exogenous.discretize()  # this is never used
        self.endo_grid = model.endo_grid
        self.gufun = gufun

    def eval_ms(self, m, s):

        return self.gufun(m, s, self.p)



def test_irf():

    from dolo import yaml_import, perturb, response
    import numpy as np

    model = yaml_import("examples/models/rbc.yaml")
    dr = perturb(model).dr

    irf = response(model, dr, "e_z")
    print(irf)


test_irf()



def test_eval_formula():

    from dolo.compiler.eval_formula import eval_formula
    from dolo import yaml_import, time_iteration, simulate
    import dolo.config

    from dolo import perturb

    model = yaml_import("examples/models/rbc.yaml")

    dr = perturb(model)
    sim = simulate(model, dr)
    sim = sim.sel(N=0)
    sim = sim.to_pandas()

    print(sim.columns)
    rr = eval_formula("delta*k(0)-i(0)", sim, context=model.calibration)
    rr = eval_formula("y(1) - y(0)", sim, context=model.calibration)

    sim["diff"] = model.eval_formula("delta*k(0)-i(0)", sim)
    model.eval_formula("y(1) - y(0)", sim)
    sim["ddiff"] = model.eval_formula("diff(1)-diff(-1)", sim)



# THis optimizer is a python port of the original LMMCP software
# written by Christian Kanzow and Stefania Petra
# Original source code and documentation can be found at:
#  http://www.mathematik.uni-wuerzburg.de/~kanzow/

# ported from lmmcp.m
import time
import numpy as np

# parameter settings
eps1 = 1e-8  # default: 1e-6
eps2 = 1e-12  # default: 1e-10
null = 1e-8  # default: 1e-8
Big = 1e10  # default: 1e+10

# maximal number of iterations
kmax = 500  # default: 500

# choice of lambda
lambda1 = 0.1  # default: 0.1
lambda2 = 1 - lambda1

# steplength parameters
beta = 0.55  # default: 0.55
sigma = 1e-4  # default: 1e-4
tmin = 1e-12  # default: 1e-12

rho = 1e-8  # default: 1e-8
p = 2.1  # default: 2.1

# parameters watchdog and nonmonotone line search redefined later
m = 10  # default: 10
kwatch = 20  # default: 20
watchdog = 1  # 1=watchdog strategy active, otherwise not

# parameters for preprocessor
preprocess = True  # 1=preprocessor used, otherwise not
presteps = 20  # maximum number of preprocessing steps, default: 20

# trust-region parameters for preprocessor
delta = 5  # default: 5
deltamin = 1  # default: 1
deltamax = 1e25  # default: 1e+25
rho1 = 1e-4  # default: rho1=1e-4
rho2 = 0.75  # default: rho2=0.75
sigma1 = 0.5  # default: 0.5
sigma2 = 2  # default: 2
eta = 0.95  # default: 0.95

# initializations
defaults = dict(
    eps1=1e-10,  # default: 1e-6
    eps2=1e-14,  # default: 1e-10
    null=1e-8,  # default: 1e-8
    Big=1e10,  # default: 1e+10
    preprocess=True,  # 1=preprocessor used, otherwise not
    presteps=20,  # maximum number of preprocessing steps, default: 20
)


def lmmcp(fun, Dfun, x0, lb, ub, verbose=True, options={}):

    # TODO: infinite bounds should be replaced by big value

    opts = dict(defaults)
    opts.update(options)

    eps1 = opts["eps1"]
    eps2 = opts["eps2"]
    null = opts["null"]
    Big = opts["Big"]
    preprocess = opts["preprocess"]
    presteps = opts["presteps"]
    # verbose = opts['verbose']

    x = x0
    delta = 5

    k = 0

    # compute a feasible starting point by projection
    x = np.maximum(lb, np.minimum(x, ub))

    n = len(x)

    # print 'Name and dimension of the testproblem: {} {}\n'.format( name, n)

    Indexset = np.zeros((n, 1))
    # I_l=find(lb>-Big & ub>Big)
    # I_u=find(lb<-Big & ub<Big)
    # I_lu=find(lb>-Big & ub<Big)
    I_l = (lb > -Big) & (ub > Big)  # not exactly the same definition
    I_u = (lb < -Big) & (ub < Big)
    I_lu = (lb > -Big) & (ub < Big)

    Indexset[I_l] = 1
    Indexset[I_u] = 2
    Indexset[I_lu] = 3

    # I_f=find(Indexset==0)
    I_f = ~(I_l | I_u | I_lu)

    # function evaluations
    Fx = fun(x)
    DFx = Dfun(x)

    # choice of NCP-function and corresponding evaluations
    # from Phi3MCPPFB import Phi3MCPPFB as Phi
    # from DPhi3MCPPFB import DPhi3MCPPFB as DPhi
    Phi = Phi3MCPPFB
    DPhi = DPhi3MCPPFB

    Phix = Phi(x, Fx, lb, ub, lambda1, lambda2, n, Indexset)

    from numpy.linalg import norm

    normPhix = norm(Phix)
    Psix = 0.5 * np.dot(Phix.T, Phix)

    DPhix = DPhi(x, Fx, DFx, lb, ub, lambda1, lambda2, n, Indexset)

    DPsix = np.dot(DPhix.T, Phix)

    normDPsix = norm(DPsix)

    # save initial values
    x0 = x
    Fx0 = Fx
    DFx0 = DFx

    Phix0 = Phix
    Psix0 = Psix
    normPhix0 = normPhix
    DPhix0 = DPhix
    DPsix0 = DPsix
    normDPsix0 = normDPsix

    # watchdog strategy

    aux = np.zeros((m, 1))
    aux[0] = Psix
    MaxPsi = Psix

    if watchdog == 1:
        kbest = k
        xbest = x
        Phibest = Phix
        Psibest = Psix
        DPhibest = DPhix
        DPsibest = DPsix
        normDPsibest = normDPsix

    if verbose:

        headline = "|{0:^5} | {1:10} | {2:12} | {3:10} |".format(
            "k", " Psi(x)", "||DPsi(x)||", "stepsize"
        )
        stars = "-" * len(headline)
        print(stars)
        print(headline)
        print(stars)
        #        print('   k               Psi(x)                || DPsi(x) ||    stepsize\n')
        #        print('====================================================================')
        s = "|{:^" + str(len(stars) - 2) + "}|"
        print(s.format("Output at starting point"))
        print(stars)
        print("|{0:5} | {1:10.3e} | {2:12.3f} |".format(k, Psix, normDPsix))

    import numpy.linalg

    if preprocess == 1:
        if verbose:
            s = "|{:^" + str(len(stars) - 2) + "}|"
            print(stars)
            print(s.format("Preprocessor"))
            print(stars)

        normpLM = 1
        while (k < presteps) & (Psix > eps2) & (normpLM > null):

            k = k + 1

            # choice of Levenberg-Marquardt parameter, note that we do not use
            # the condition estimator for large-scale problems, although this
            # may cause numerical problems in some examples

            i = 0
            mu = 0
            if n < 100:
                i = 1
                mu = 1e-16
                if (
                    numpy.linalg.cond(np.dot(DPhix.T, DPhix)) > 1e25
                ):  # TODO use random estimator
                    mu = 1e-6 / (k + 1)

            if i == 1:
                A1 = np.row_stack([DPhix, np.sqrt(mu) * np.eye(n)])
                A2 = np.concatenate([-Phix, np.zeros(n)])
                #            pLM = A1 \ A2
                pLM = np.linalg.lstsq(A1, A2, rcond=None)[0]
            else:
                # pLM = A1 \ A2
                pLM = -np.linalg.lstsq(DPhix, Phix)[0]

            normpLM = norm(pLM)

            pLM = pLM.flatten()

            # compute the projected Levenberg-Marquard step onto box Xk
            lbnew = np.maximum(np.minimum(lb - x, 0), -delta)
            ubnew = np.minimum(np.maximum(ub - x, 0), delta)
            d = np.maximum(lbnew, np.minimum(pLM, ubnew))
            xnew = x + d

            # function evaluations etc.
            Fxnew = fun(xnew)
            DFxnew = Dfun(xnew)
            Phixnew = Phi(xnew, Fxnew, lb, ub, lambda1, lambda2, n, Indexset)
            Psixnew = 0.5 * np.dot(Phixnew.T, Phixnew)
            normPhixnew = norm(Phixnew)

            # update of delta
            if normPhixnew <= eta * normPhix:
                delta = max(deltamin, sigma2 * delta)
            elif normPhixnew > 5 * eta * normPhix:
                delta = max(deltamin, sigma1 * delta)

            # update
            x = xnew
            Fx = Fxnew
            DFx = DFxnew
            Phix = Phixnew
            Psix = Psixnew
            normPhix = normPhixnew
            DPhix = DPhi(x, Fx, DFx, lb, ub, lambda1, lambda2, n, Indexset)
            DPsix = np.dot(DPhix.T, Phix)
            normDPsix = norm(DPsix, ord=np.inf)

            # output at each iteration
            t = 1
            if verbose:
                print(
                    "|{0:5} | {1:10.3e} | {2:12.3f} | {3:10.3f} |".format(
                        k, Psix, normDPsix, t
                    )
                )

    #                print('{}\t{}\t{}\t{}\n'.format(k,Psix,normDPsix,t))

    # terminate program or redefine current iterate as original initial point
    if (preprocess == 1) & (Psix < eps2):
        if verbose:
            s = "|{:^" + str(len(stars) - 2) + "}|"
            print(stars)
            print(s.format("Approximate solution found"))
            print(stars)
            return x

    elif (preprocess == 1) & (Psix >= eps2):
        k_main = 0

        x = x0
        Fx = Fx0
        Phix = Phix0
        Psix = Psix0
        normPhix = normPhix0
        DPhix = DPhix0
        DPsix = DPsix0
        normDPsi = normDPsix0
        if verbose:
            s = "|{:^" + str(len(stars) - 2) + "}|"
            print(s.format("Restart with initial point"))
            print("{}\t{}\t{}\n".format(k_main, Psix0, normDPsix0))

    #
    #   Main algorithm
    #
    if verbose:
        s = "|{:^" + str(len(stars) - 2) + "}|"
        print(stars)
        print(s.format("Main program"))
        print(stars)

    k_main = 0

    while (k < kmax) & (Psix > eps2):

        # choice of Levenberg-Marquardt parameter, note that we do not use
        # the condition estimator for large-scale problems, although this
        # may cause numerical problems in some examples

        i = 0
        if n < 100:
            i = 1
            mu = 1e-16
            if (
                numpy.linalg.cond(np.dot(DPhix.T, DPhix)) > 1e25
            ):  # TODO use random estimator
                mu = 1e-1 / (k + 1)
        if i == 1:
            A1 = np.row_stack([DPhix, np.sqrt(mu) * np.eye(n)])
            A2 = np.concatenate([-Phix, np.zeros(n)])
            #            pLM = A1 \ A2
            d = np.linalg.lstsq(A1, A2)[0]
        else:
            # pLM = A1 \ A2
            d = -np.linalg.lstsq(DPhix, Phix)[0]

        d = d.flatten()

        # computation of steplength t using the nonmonotone Armijo-rule
        # starting with the 6-th iteration

        # computation of steplength t using the monotone Armijo-rule if
        # d is a 'good' descent direction or k<=5

        t = 1
        xnew = x + d
        Fxnew = fun(xnew)
        Phixnew = Phi(xnew, Fxnew, lb, ub, lambda1, lambda2, n, Indexset)
        Psixnew = 0.5 * np.dot(Phixnew.T, Phixnew)
        const = np.dot(sigma * DPsix.T, d)

        while (Psixnew > MaxPsi + const * t) & (t > tmin):
            t = t * beta
            xnew = x + t * d
            Fxnew = fun(xnew)
            Phixnew = Phi(xnew, Fxnew, lb, ub, lambda1, lambda2, n, Indexset)
            Psixnew = 0.5 * np.dot(Phixnew.T, Phixnew)

        # updatings
        x = xnew
        Fx = Fxnew
        Phix = Phixnew
        Psix = Psixnew
        DFx = Dfun(x)
        DPhix = DPhi(x, Fx, DFx, lb, ub, lambda1, lambda2, n, Indexset)
        DPsix = np.dot(DPhix.T, Phix)
        normDPsix = norm(DPsix)
        k = k + 1

        k_main = k_main + 1

        if k_main <= 5:
            aux[numpy.mod(k_main, m)] = Psix
            MaxPsi = Psix
        else:
            aux[numpy.mod(k_main, m)] = Psix
            MaxPsi = max(aux)

        # updatings for the watchdog strategy
        if watchdog == 1:
            if Psix < Psibest:
                kbest = k
                xbest = x
                Phibest = Phix
                Psibest = Psix
                DPhibest = DPhix
                DPsibest = DPsix
                normDPsibest = normDPsix
            elif k - kbest > kwatch:
                x = xbest
                Phix = Phibest
                Psix = Psibest
                DPhix = DPhibest
                DPsix = DPsibest
                normDPsix = normDPsibest
                MaxPsi = Psix
        # output at each iteration
        if verbose:
            print("{}\t{}\t{}\t{}\n".format(k, Psix, normDPsix, t))

    if k == kmax:
        raise Exception("No convergence")
    return x


################### helper functions #######################


def Phi3MCPPFB(x, Fx, lb, ub, lambda1, lambda2, n, Indexset):
    y = np.zeros(2 * n)
    for i in range(1, n + 1):
        phi_u = (
            np.sqrt((ub[i - 1] - x[i - 1]) ** 2 + Fx[i - 1] ** 2)
            - ub[i - 1]
            + x[i - 1]
            + Fx[i - 1]
        )
        if Indexset[i - 1] == 1:
            y[i - 1] = lambda1 * (
                -x[i - 1]
                + lb[i - 1]
                - Fx[i - 1]
                + np.sqrt((x[i - 1] - lb[i - 1]) ** 2 + Fx[i - 1] ** 2)
            )
            y[n + i - 1] = lambda2 * max(0, x[i - 1] - lb[i - 1]) * max(0, Fx[i - 1])
        elif Indexset[i - 1] == 2:
            y[i - 1] = -lambda1 * phi_u
            y[n + i - 1] = lambda2 * max(0, ub[i - 1] - x[i - 1]) * max(0, -Fx[i - 1])
        elif Indexset[i - 1] == 0:
            y[i - 1] = -lambda1 * Fx[i - 1]
            y[n + i - 1] = -lambda2 * Fx[i - 1]
        elif Indexset[i - 1] == 3:
            y[i - 1] = lambda1 * (
                np.sqrt((x[i - 1] - lb[i - 1]) ** 2 + phi_u ** 2)
                - x[i - 1]
                + lb[i - 1]
                - phi_u
            )
            y[n + i - 1] = lambda2 * (
                max(0, x[i - 1] - lb[i - 1]) * max(0, Fx[i - 1])
                + max(0, ub[i - 1] - x[i - 1]) * max(0, -Fx[i - 1])
            )
    return y


def DPhi3MCPPFB(x, Fx, DFx, lb, ub, lambda1, lambda2, n, Indexset):

    #% we evaluate an element of the C-subdifferential of operator Phi3MCPPFB
    null = 1e-8
    beta_l = np.zeros(n)
    beta_u = np.zeros(n)
    alpha_l = np.zeros(n)
    alpha_u = np.zeros(n)
    z = np.zeros(n)
    H1 = np.zeros((n, n))  # should be sparse
    H2 = H1.copy()
    for i in range(1, n + 1):
        if np.logical_and(
            np.abs((x[i - 1] - lb[i - 1])) <= null, np.abs(Fx[i - 1]) <= null
        ):
            beta_l[i - 1] = 1.0
            z[i - 1] = 1.0

        if np.logical_and(
            np.abs((ub[i - 1] - x[i - 1])) <= null, np.abs(Fx[i - 1]) <= null
        ):
            beta_u[i - 1] = 1.0
            z[i - 1] = 1.0

        if np.logical_and(x[i - 1] - lb[i - 1] >= -null, Fx[i - 1] >= -null):
            alpha_l[i - 1] = 1.0

        if np.logical_and(ub[i - 1] - x[i - 1] >= -null, Fx[i - 1] <= null):
            alpha_u[i - 1] = 1.0

    Da = np.zeros((n, 1))
    Db = np.zeros((n, 1))
    for i in range(1, n + 1):
        #        ei = np.zeros( (1., n) )
        ei = np.zeros(n)
        ei[i - 1] = 1.0
        if Indexset[i - 1] == 0.0:
            Da[i - 1] = 0
            Db[i - 1] = -1
            H2[i - 1, :] = -DFx[i - 1, :]
        elif Indexset[i - 1] == 1.0:
            # TODO : hyp : the maximand is a scalar
            denom1 = np.maximum(
                null, np.sqrt(((x[i - 1] - lb[i - 1]) ** 2 + Fx[i - 1] ** 2))
            )
            denom2 = np.maximum(
                null, np.sqrt((z[i - 1] ** 2 + np.dot(DFx[i - 1, :], z) ** 2))
            )
            if beta_l[i - 1] == 0.0:
                Da[i - 1] = (x[i - 1] - lb[i - 1]) / denom1 - 1.0
                Db[i - 1] = Fx[i - 1] / denom1 - 1.0
            else:
                Da[i - 1] = z[i - 1] / denom2 - 1.0
                Db[i - 1] = np.dot(DFx[i - 1, :], z) / denom2 - 1.0

            if alpha_l[i - 1] == 1.0:
                H2[i - 1, :] = (x[i - 1] - lb[i - 1]) * DFx[i - 1, :] + Fx[i - 1] * ei
            else:
                H2[i - 1, :] = 0.0

        elif Indexset[i - 1] == 2:
            denom1 = np.maximum(
                null, np.sqrt(((ub[i - 1] - x[i - 1]) ** 2 + Fx[i - 1] ** 2))
            )
            denom2 = np.maximum(
                null, np.sqrt((z[i - 1] ** 2 + np.dot(DFx[i - 1, :], z) ** 2))
            )
            if beta_u[i - 1] == 0.0:
                Da[i - 1] = (ub[i - 1] - x[i - 1]) / denom1 - 1.0
                Db[i - 1] = -Fx[i - 1] / denom1 - 1.0
            else:
                Da[i - 1] = -z[i - 1] / denom2 - 1.0
                Db[i - 1] = -np.dot(DFx[i - 1, :], z) / denom2 - 1.0

            if alpha_u[i - 1] == 1.0:
                H2[i - 1, :] = np.dot(x[i - 1] - ub[i - 1], DFx[i - 1, :]) + np.dot(
                    Fx[i - 1], ei
                )
            else:
                H2[i - 1, :] = 0.0

        elif Indexset[i - 1] == 3.0:
            ai = 0.0
            bi = 0.0
            ci = 0.0
            di = 0.0
            phi = (
                -ub[i - 1]
                + x[i - 1]
                + Fx[i - 1]
                + np.sqrt(((ub[i - 1] - x[i - 1]) ** 2 + Fx[i - 1] ** 2))
            )
            denom1 = np.maximum(null, np.sqrt(((x[i - 1] - lb[i - 1]) ** 2 + phi ** 2)))
            denom2 = np.maximum(
                null, np.sqrt((z[i - 1] ** 2 + np.dot(DFx[i - 1, :], z) ** 2))
            )
            denom3 = np.maximum(
                null, np.sqrt(((ub[i - 1] - x[i - 1]) ** 2 + Fx[i - 1] ** 2))
            )
            denom4 = np.maximum(
                null,
                np.sqrt(
                    (
                        z[i - 1] ** 2
                        + (np.dot(ci, z[i - 1]) + np.dot(np.dot(di, DFx[i - 1, :]), z))
                        ** 2
                    )
                ),
            )

            if beta_u[i - 1] == 0.0:
                ci = (x[i - 1] - ub[i - 1]) / denom3 + 1
                di = Fx[i - 1] / denom3 + 1
            else:
                ci = 1 + z[i - 1] / denom2
                di = 1 + np.dot(DFx[i - 1, :], z) / denom2

            if beta_l[i - 1] == 0.0:
                ai = (x[i - 1] - lb[i - 1]) / denom1 - 1
                bi = phi / denom1 - 1
            else:
                ai = z[i - 1] / denom4 - 1.0
                bi = (
                    np.dot(ci, z[i - 1]) + np.dot(np.dot(di, DFx[i - 1, :]), z)
                ) / denom4 - 1.0

            Da[i - 1] = ai + np.dot(bi, ci)
            Db[i - 1] = np.dot(bi, di)
            if np.logical_and(alpha_l[i - 1] == 1.0, alpha_u[i - 1] == 1.0):
                H2[i - 1, :] = np.dot(
                    -lb[i - 1] - ub[i - 1] + 2 * x[i - 1], DFx[i - 1, :]
                ) + np.dot(2 * Fx[i - 1], ei)
            else:
                if alpha_l[i - 1] == 1.0:
                    H2[i - 1, :] = np.dot(x[i - 1] - lb[i - 1], DFx[i - 1, :]) + np.dot(
                        Fx[i - 1], ei
                    )
                elif alpha_u[i - 1] == 1.0:
                    H2[i - 1, :] = np.dot(x[i - 1] - ub[i - 1], DFx[i - 1, :]) + np.dot(
                        Fx[i - 1], ei
                    )

                else:
                    H2[i - 1, :] = 0.0

        H1[i - 1, :] = Da[i - 1] * ei + Db[i - 1] * DFx[i - 1, :]

    #  H=[lambda1*H1; lambda2*H2];
    H = np.row_stack([lambda1 * H1, lambda2 * H2])
    #    H = np.array(np.vstack((np.hstack((np.dot(lambda1, H1))), np.hstack((np.dot(lambda2, H2))))))

    return H



from dolo.numeric.distribution import *

#
from dataclasses import dataclass
from dolang.language import language_element

# not sure we'll keep that
import numpy as np
from typing import List, Union

Scalar = Union[int, float]

# not really a language element though
# @language_element


class Domain:
    pass


class CartesianDomain(Domain, dict):
    def __init__(self, **kwargs):
        super().__init__()
        for k, w in kwargs.items():
            v = kwargs[k]
            self[k] = np.array(v, dtype=float)

    def discretize(self, n=None):
        if n == None:
            n = [10] * (len(self.min))
        from dolo.numeric.grids import UniformCartesianGrid

        return UniformCartesianGrid(self.min, self.max, n)

    @property
    def states(self):
        l = tuple(self.keys())
        return l

    @property
    def min(self):
        return np.array([self[e][0] for e in self.states])

    @property
    def max(self):
        return np.array([self[e][1] for e in self.states])


# these are dummy objects so far
#
# @language_element
# @dataclass
# class UNormal:
#     mu: float
#     sigma: float
#     signature = {'mu': 'float', 'sigma': 'float'}
#


# @language_element
# @dataclass
# class MvNormal:
#     Mu: List[float]
#     Sigma: List[List[float]]
#     signature = {'Mu': 'list(float)', 'Sigma': 'Matrix'}


#%%


@language_element
class Conditional:

    signature = {"condition": None, "type": None, "arguments": None}

    def __init__(self, condition, type, arguments):
        self.condition = condition
        self.type = type
        self.arguments = arguments


@language_element
class Product:
    def __init__(self, *args: List):
        self.factors = args


@language_element
def Matrix(*lines):
    mat = np.array(lines, np.float64)
    assert mat.ndim == 2
    return mat


@language_element
def Vector(*elements):
    mat = np.array(elements, np.float64)
    assert mat.ndim == 1
    return mat



import numpy

from dolo.misc.display import read_file_or_url
import yaml


def yaml_import(fname, check=True, check_only=False):

    txt = read_file_or_url(fname)

    try:
        data = yaml.compose(txt)
        # print(data)
        # return data
    except Exception as ex:
        print(
            "Error while parsing YAML file. Probable YAML syntax error in file : ",
            fname,
        )
        raise ex

    # if check:
    #     from dolo.linter import lint
    #     data = ry.load(txt, ry.RoundTripLoader)
    #     output = lint(data, source=fname)
    #     if len(output) > 0:
    #         print(output)

    # if check_only:
    #     return output

    data["filename"] = fname

    from dolo.compiler.model import Model

    return Model(data, check=check)
    # from dolo.compiler.model import SymbolicModel



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dolo',
 'dolo.algos',
 'dolo.compiler',
 'dolo.misc',
 'dolo.numeric',
 'dolo.numeric.discretization',
 'dolo.numeric.extern',
 'dolo.numeric.optimize',
 'dolo.tests']

package_data = \
{'': ['*']}

install_requires = \
['dolang==0.0.14',
 'interpolation>=2.1.6,<3.0.0',
 'ipython>=7.13.0,<8.0.0',
 'matplotlib>=3.0.1,<4.0.0',
 'multipledispatch>=0.6.0,<0.7.0',
 'numpy>=1.16.0,<2.0.0',
 'pandas>=1.0.3,<2.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'quantecon>=0.4.7,<0.5.0',
 'scipy>=1.4.1,<2.0.0',
 'xarray>=0.15.1,<0.16.0']

setup_kwargs = {
    'name': 'dolo',
    'version': '0.4.9.15',
    'description': 'Economic Modeling in Python',
    'long_description': 'Complete documentation with installation instruction, available at http://dolo.readthedocs.org/en/latest/\nNew in-progress [version](https://www.econforge.org/dolo.py).\n\nJoin the chat at https://gitter.im/EconForge/dolo\n\n[![codecov](https://codecov.io/gh/EconForge/dolo.py/branch/master/graph/badge.svg?token=hLAd1OaTRp)](https://codecov.io/gh/EconForge/dolo.py)\n\n![CI](https://github.com/EconForge/dolo.py/workflows/CI/badge.svg)\n\n![Publish docs via GitHub Pages](https://github.com/EconForge/dolo.py/workflows/Publish%20docs%20via%20GitHub%20Pages/badge.svg)\n\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/EconForge/dolo.git/master?urlpath=lab)\n',
    'author': 'Winant Pablo',
    'author_email': 'pablo.winant@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



def test_grids():

    from dolo.numeric.grids import (
        UniformCartesianGrid,
        UnstructuredGrid,
        NonUniformCartesianGrid,
        SmolyakGrid,
    )
    from dolo.numeric.grids import nodes, n_nodes, node

    print("Cartsian Grid")
    grid = UniformCartesianGrid([0.1, 0.3], [9, 0.4], [50, 10])
    print(grid.nodes)
    print(nodes(grid))

    print("UnstructuredGrid")
    ugrid = UnstructuredGrid([[0.1, 0.3], [9, 0.4], [50, 10]])
    print(nodes(ugrid))
    print(node(ugrid, 0))
    print(n_nodes(ugrid))

    print("Non Uniform CartesianGrid")
    ugrid = NonUniformCartesianGrid([[0.1, 0.3], [9, 0.4], [50, 10]])
    print(nodes(ugrid))
    print(node(ugrid, 0))
    print(n_nodes(ugrid))

    print("Smolyak Grid")
    sg = SmolyakGrid([0.1, 0.2], [1.0, 2.0], 2)
    print(nodes(sg))
    print(node(sg, 1))
    print(n_nodes(sg))



import time
import numpy as np
import numpy
import scipy.optimize
from dolo.numeric.processes import DiscretizedIIDProcess

# from dolo.numeric.decision_rules_markov import MarkovDecisionRule, IIDDecisionRule
from dolo.numeric.decision_rule import DecisionRule, ConstantDecisionRule
from dolo.numeric.grids import Grid, CartesianGrid, SmolyakGrid, UnstructuredGrid
from dolo.misc.itprinter import IterationsPrinter


def constant_policy(model):
    return ConstantDecisionRule(model.calibration["controls"])


from .results import AlgoResult, ValueIterationResult


def value_iteration(
    model, tol=1e-6, maxit=500, maxit_howard=20, verbose=False, details=True
):
    """
    Solve for the value function and associated Markov decision rule by iterating over
    the value function.

    Parameters:
    -----------
    model :
        "dtmscc" model. Must contain a 'felicity' function.
    dr :
        decision rule to evaluate

    Returns:
    --------
    mdr : Markov decision rule
        The solved decision rule/policy function
    mdrv: decision rule
        The solved value function
    """

    transition = model.functions["transition"]
    felicity = model.functions["felicity"]
    controls_lb = model.functions["controls_lb"]
    controls_ub = model.functions["controls_ub"]

    parms = model.calibration["parameters"]
    discount = model.calibration["beta"]

    x0 = model.calibration["controls"]
    m0 = model.calibration["exogenous"]
    s0 = model.calibration["states"]
    r0 = felicity(m0, s0, x0, parms)

    process = model.exogenous

    grid, dprocess = model.discretize()
    endo_grid = grid["endo"]
    exo_grid = grid["exo"]

    n_ms = dprocess.n_nodes  # number of exogenous states
    n_mv = dprocess.n_inodes(0)  # this assume number of integration nodes is constant

    mdrv = DecisionRule(exo_grid, endo_grid)

    s = mdrv.endo_grid.nodes
    N = s.shape[0]
    n_x = len(x0)

    mdr = constant_policy(model)

    controls_0 = np.zeros((n_ms, N, n_x))
    for i_ms in range(n_ms):
        controls_0[i_ms, :, :] = mdr.eval_is(i_ms, s)

    values_0 = np.zeros((n_ms, N, 1))
    # for i_ms in range(n_ms):
    #     values_0[i_ms, :, :] = mdrv(i_ms, grid)

    mdr = DecisionRule(exo_grid, endo_grid)
    # mdr.set_values(controls_0)

    # THIRD: value function iterations until convergence
    it = 0
    err_v = 100
    err_v_0 = 0
    gain_v = 0.0
    err_x = 100
    err_x_0 = 0
    tol_x = 1e-5
    tol_v = 1e-7

    itprint = IterationsPrinter(
        ("N", int),
        ("Error_V", float),
        ("Gain_V", float),
        ("Error_x", float),
        ("Gain_x", float),
        ("Eval_n", int),
        ("Time", float),
        verbose=verbose,
    )
    itprint.print_header("Start value function iterations.")

    while (it < maxit) and (err_v > tol or err_x > tol_x):

        t_start = time.time()
        it += 1

        mdr.set_values(controls_0)
        if it > 2:
            ev = evaluate_policy(model, mdr, dr0=mdrv, verbose=False, details=True)
        else:
            ev = evaluate_policy(model, mdr, verbose=False, details=True)

        mdrv = ev.solution
        for i_ms in range(n_ms):
            values_0[i_ms, :, :] = mdrv.eval_is(i_ms, s)

        values = values_0.copy()
        controls = controls_0.copy()

        for i_m in range(n_ms):
            m = dprocess.node(i_m)
            for n in range(N):
                s_ = s[n, :]
                x = controls[i_m, n, :]
                lb = controls_lb(m, s_, parms)
                ub = controls_ub(m, s_, parms)
                bnds = [e for e in zip(lb, ub)]

                def valfun(xx):
                    return -choice_value(
                        transition,
                        felicity,
                        i_m,
                        s_,
                        xx,
                        mdrv,
                        dprocess,
                        parms,
                        discount,
                    )[0]

                res = scipy.optimize.minimize(valfun, x, bounds=bnds)
                controls[i_m, n, :] = res.x
                values[i_m, n, 0] = -valfun(x)

        # compute error, update value and dr
        err_x = abs(controls - controls_0).max()
        err_v = abs(values - values_0).max()
        t_end = time.time()
        elapsed = t_end - t_start

        values_0 = values
        controls_0 = controls

        gain_x = err_x / err_x_0
        gain_v = err_v / err_v_0

        err_x_0 = err_x
        err_v_0 = err_v

        itprint.print_iteration(
            N=it,
            Error_V=err_v,
            Gain_V=gain_v,
            Error_x=err_x,
            Gain_x=gain_x,
            Eval_n=ev.iterations,
            Time=elapsed,
        )

    itprint.print_finished()

    mdr = DecisionRule(exo_grid, endo_grid)

    mdr.set_values(controls)
    mdrv.set_values(values_0)

    if not details:
        return mdr, mdrv
    else:
        return ValueIterationResult(
            mdr,  #:AbstractDecisionRule
            mdrv,  #:AbstractDecisionRule
            it,  #:Int
            dprocess,  #:AbstractDiscretizedProcess
            err_x < tol_x,  #:Bool
            tol_x,  #:Float64
            err_x,  #:Float64
            err_v < tol_v,  #:Bool
            tol_v,  #:Float64
            err_v,  #:Float64
            None,  # log:     #:ValueIterationLog
            None,  # trace:   #:Union{Nothing,IterationTrace
        )


def choice_value(transition, felicity, i_ms, s, x, drv, dprocess, parms, beta):

    m = dprocess.node(i_ms)
    cont_v = 0.0
    for I_ms in range(dprocess.n_inodes(i_ms)):
        M = dprocess.inode(i_ms, I_ms)
        prob = dprocess.iweight(i_ms, I_ms)
        S = transition(m, s, x, M, parms)
        V = drv(I_ms, S)[0]
        cont_v += prob * V
    return felicity(m, s, x, parms) + beta * cont_v


class EvaluationResult:
    def __init__(self, solution, iterations, tol, error):
        self.solution = solution
        self.iterations = iterations
        self.tol = tol
        self.error = error


def evaluate_policy(
    model,
    mdr,
    tol=1e-8,
    maxit=2000,
    grid={},
    verbose=True,
    dr0=None,
    hook=None,
    integration_orders=None,
    details=False,
    interp_method="cubic",
):
    """Compute value function corresponding to policy ``dr``

    Parameters:
    -----------

    model:
        "dtcscc" model. Must contain a 'value' function.

    mdr:
        decision rule to evaluate

    Returns:
    --------

    decision rule:
        value function (a function of the space similar to a decision rule
        object)

    """

    process = model.exogenous
    grid, dprocess = model.discretize()
    endo_grid = grid["endo"]
    exo_grid = grid["exo"]

    n_ms = dprocess.n_nodes  # number of exogenous states
    n_mv = dprocess.n_inodes(0)  # this assume number of integration nodes is constant

    x0 = model.calibration["controls"]
    v0 = model.calibration["values"]
    parms = model.calibration["parameters"]
    n_x = len(x0)
    n_v = len(v0)
    n_s = len(model.symbols["states"])

    if dr0 is not None:
        mdrv = dr0
    else:
        mdrv = DecisionRule(exo_grid, endo_grid, interp_method=interp_method)

    s = mdrv.endo_grid.nodes
    N = s.shape[0]

    if isinstance(mdr, np.ndarray):
        controls = mdr
    else:
        controls = np.zeros((n_ms, N, n_x))
        for i_m in range(n_ms):
            controls[i_m, :, :] = mdr.eval_is(i_m, s)

    values_0 = np.zeros((n_ms, N, n_v))
    if dr0 is None:
        for i_m in range(n_ms):
            values_0[i_m, :, :] = v0[None, :]
    else:
        for i_m in range(n_ms):
            values_0[i_m, :, :] = dr0.eval_is(i_m, s)

    val = model.functions["value"]
    g = model.functions["transition"]

    sh_v = values_0.shape

    err = 10
    inner_maxit = 50
    it = 0

    if verbose:
        headline = "|{0:^4} | {1:10} | {2:8} | {3:8} |".format(
            "N", " Error", "Gain", "Time"
        )
        stars = "-" * len(headline)
        print(stars)
        print(headline)
        print(stars)

    t1 = time.time()

    err_0 = np.nan

    verbit = verbose == "full"

    while err > tol and it < maxit:

        it += 1

        t_start = time.time()

        mdrv.set_values(values_0.reshape(sh_v))
        values = update_value(
            val, g, s, controls, values_0, mdr, mdrv, dprocess, parms
        ).reshape((-1, n_v))
        err = abs(values.reshape(sh_v) - values_0).max()

        err_SA = err / err_0
        err_0 = err

        values_0 = values.reshape(sh_v)

        t_finish = time.time()
        elapsed = t_finish - t_start

        if verbose:
            print(
                "|{0:4} | {1:10.3e} | {2:8.3f} | {3:8.3f} |".format(
                    it, err, err_SA, elapsed
                )
            )

    # values_0 = values.reshape(sh_v)

    t2 = time.time()

    if verbose:
        print(stars)
        print("Elapsed: {} seconds.".format(t2 - t1))
        print(stars)

    if not details:
        return mdrv
    else:
        return EvaluationResult(mdrv, it, tol, err)


def update_value(val, g, s, x, v, dr, drv, dprocess, parms):

    N = s.shape[0]
    n_s = s.shape[1]

    n_ms = dprocess.n_nodes  # number of exogenous states
    n_mv = dprocess.n_inodes(0)  # this assume number of integration nodes is constant

    res = np.zeros_like(v)

    for i_ms in range(n_ms):

        m = dprocess.node(i_ms)[None, :].repeat(N, axis=0)

        xm = x[i_ms, :, :]
        vm = v[i_ms, :, :]

        for I_ms in range(n_mv):

            # M = P[I_ms,:][None,:]
            M = dprocess.inode(i_ms, I_ms)[None, :].repeat(N, axis=0)
            prob = dprocess.iweight(i_ms, I_ms)

            S = g(m, s, xm, M, parms)
            XM = dr.eval_ijs(i_ms, I_ms, S)
            VM = drv.eval_ijs(i_ms, I_ms, S)
            rr = val(m, s, xm, vm, M, S, XM, VM, parms)

            res[i_ms, :, :] += prob * rr

    return res



import numpy as np


def multidot_old(ten, mats):
    """
    Implements tensor operation : tensor-times-matrices.
    If last dimensions of ten represent multilinear operations of the type : [X1,...,Xk]->B[X1,...,Xk]
    and mats contains matrices or vectors [A1,...Ak] the function returns an array representing operators :
    [X1,...,Xk]->B[A1 X1,...,Ak Xk]
    """
    resp = ten
    n_d = ten.ndim
    n_m = len(mats)
    for i in range(n_m):
        # resp = np.tensordot( resp, mats[i], (n_d-n_m+i-1,0) )
        resp = np.tensordot(resp, mats[i], (n_d - n_m, 0))
    return resp


# should consist in shapes only
def mdot_signature(M, *C):
    M_syms = [chr(97 + e) for e in range(len(M))]
    fC_syms = M_syms[-len(C) :]
    ic = 97 + len(M_syms)
    C_syms = []
    for i in range(len(C)):
        c_sym = [fC_syms[i]]
        for j in range(len(C[i]) - 1):
            c_sym.append(chr(ic))
            ic += 1
        C_syms.append(c_sym)
    C_sig = [M_syms] + C_syms
    out_sig = [M_syms[: -len(C)]] + [cc[1:] for cc in C_syms]
    args = ",".join(["".join(g) for g in C_sig])
    out = "".join(["".join(g) for g in out_sig])

    return args + "->" + out


from numpy import einsum


def mdot(M, *C):
    sig = mdot_signature(M.shape, *[c.shape for c in C])
    return einsum(sig, M, *C)


def sdot(U, V):
    """
    Computes the tensorproduct reducing last dimensoin of U with first dimension of V.
    For matrices, it is equal to regular matrix product.
    """
    nu = U.ndim
    # nv = V.ndim
    return np.tensordot(U, V, axes=(nu - 1, 0))


def multitake(a, inds, axes):
    resp = a
    for n, ind in enumerate(inds):
        axe = axes[n]
        resp = resp.take(ind, axe)
    return resp



def test_custom_dr():

    from dolo import yaml_import, simulate, time_iteration
    import numpy as np
    from dolo.numeric.decision_rule import CustomDR

    model = yaml_import("examples/models/rbc.yaml")

    values = {"n": "0.33 + z*0.01", "i": "delta*k-0.07*(k-9.35497829)"}

    edr = CustomDR(values, model)

    m0, s0 = model.calibration["exogenous", "controls"]

    edr(m0, s0)

    sim = simulate(model, edr, s0=np.array([0.0, 8.0]))

    time_iteration(model, dr0=edr)



import inspect


class bcolors:

    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"

    def disable(self):
        self.HEADER = ""
        self.OKBLUE = ""
        self.OKGREEN = ""
        self.WARNING = ""
        self.FAIL = ""
        self.ENDC = ""


def dprint(s):
    """Prints `s` with additional debugging informations"""

    import inspect

    frameinfo = inspect.stack()[1]
    callerframe = frameinfo.frame
    d = callerframe.f_locals

    if isinstance(s, str):
        val = eval(s, d)
    else:
        val = s
        cc = frameinfo.code_context[0]
        import re

        regex = re.compile("dprint\((.*)\)")
        res = regex.search(cc)
        s = res.group(1)

    text = ""
    text += bcolors.OKBLUE + "At <{}>\n".format(str(frameinfo)) + bcolors.ENDC
    text += bcolors.WARNING + "{}:  ".format(s) + bcolors.ENDC
    text += str(val)
    text += str()

    print(text)


if __name__ == "__main__":

    a = 34

    dprint("a")

    dprint(a)



"""
ideas :
-  recursive blocks           [by default]
- (order left hand side ?)    [by default]
- dependency across blocks
- dummy blocks that are basically substituted everywhere else
"""

import os, yaml, sys

if getattr(sys, "frozen", False):
    # we are running in a |PyInstaller| bundle
    DIR_PATH = sys._MEIPASS
else:
    DIR_PATH, this_filename = os.path.split(__file__)

DATA_PATH = os.path.join(DIR_PATH, "recipes.yaml")

with open(DATA_PATH) as f:
    recipes = yaml.safe_load(f)



import numpy as np
import numpy
import pickle

from typing import List, TypeVar, Generic, Union, Any, Callable  # type: ignore
from typing import Iterator, Tuple  # type: ignore
from typing import List, Optional

Vector = List[float]
Matrix = List[Vector]

from numba import jit, njit
from dolo.numeric.grids import EmptyGrid, CartesianGrid, UnstructuredGrid
from random import random as rand

from xarray import DataArray

from dolang.language import greek_tolerance
from dolang.language import language_element


@njit
def choice(x, n, cumul):
    i = 0
    running = True
    while i < n and running:
        if x < cumul[i]:
            running = False
        else:
            i += 1
    return i


@njit
def simulate_mc_indices(C, N, T, i0):
    k = len(C)
    out = np.zeros((T, N), dtype=np.int32)
    out[0, :] = i0
    for n in range(N):
        for t in range(0, T - 1):
            ii = out[t, n]
            options = C[ii]
            x = rand()
            ix = choice(x, k, options)
            out[t + 1, n] = ix
    return out


def simulate_markov_chain(nodes, transitions, N, T, i0=None, return_values=False):
    i0_ = np.zeros(N, dtype=np.int32)
    if i0 is not None:
        i0_[:] = i0
    C = transitions.cumsum(axis=1)
    indices = simulate_mc_indices(C, N, T, i0_)
    if return_values:
        values = np.reshape(
            np.concatenate([Q[i, :][None, :] for i in indices.ravel()], axis=0),
            indices.shape + (-1,),
        )
        return indices, values
    else:
        return indices


class Process:

    d: int  # number of dimensions
    variables: Union[None, Tuple[str, ...]]  # names of variables (optional)

    @property
    def names(self):
        try:
            return self._names_
        except:
            self._names_ = tuple(f"_x_{i}" for i in range(self.d))
            return self._names_

    @names.setter
    def names(self, v):
        self._names_ = v


# A discrete process is such that integration
# can be computed exactly.
class DiscreteProcess(Process):

    n: int


class ContinuousProcess(Process):
    pass


class IIDProcess(Process):
    def draw(self, n: int) -> Vector:
        raise Exception("Not Implemented")

    def response(self, T, impulse) -> DataArray:
        irf = numpy.zeros((T + 1, self.d))
        irf[1, :] = impulse[None, :]
        coords = {"T": range(T + 1), "V": list(self.names)}
        xar = DataArray(irf, dims=coords.keys(), coords=coords)
        return xar

    def simulate(
        self, N: int = 10, T: int = 100, i0=None, m0=None, stochastic=True
    ) -> DataArray:
        sim = self.draw((T) * N).reshape((T, N, -1))
        coords = {"T": range(T), "N": range(N), "V": list(self.names)}
        xar = DataArray(sim, dims=coords.keys(), coords=coords)
        return xar


@language_element
class ConstantProcess(IIDProcess):

    signature = {"μ": "Vector"}

    @greek_tolerance
    def __init__(self, μ=None):

        if isinstance(μ, (float, int)):
            μ = [μ]
        self.μ = np.array(μ)
        assert self.μ.ndim == 1
        self.d = len(self.μ)

    def discretize(self, to=None, **kwargs):

        if to == "iid":
            x = self.μ[None, :]
            w = np.array([1.0])
            from .distribution import FiniteDistribution

            return FiniteDistribution(x, w)

        elif to == "mc":
            N = kwargs.get("N", 1)
            nodes = self.μ[None, :].repeat(N, axis=0)
            transitions = np.eye(N)
            return MarkovChain(transitions, nodes)
        else:
            raise Exception("Not implemented")


@language_element
class AggregateProcess(ConstantProcess):

    # that is a dummy class
    pass


# class VAR1(ContinuousProcess):
#     pass


class DiscretizedProcess:
    def iweight(self, i: int, j: int) -> float:
        raise Exception("Not Implemented.")

    def inode(self, i: int, j: int) -> float:
        raise Exception("Not Implemented.")

    def iteritems(self, i, eps=1e-16):
        for j in range(self.n_inodes(i)):
            w = self.iweight(i, j)
            if w > eps:
                x = self.inode(i, j)
                yield (w, x)


class GDP(DiscretizedProcess):
    def __init__(self, nodes, inodes, iweights, grid=None):
        self.__nodes__ = nodes
        self.__inodes__ = inodes
        self.__iweights__ = iweights
        self.__grid__ = grid

    # def discretize_gdp(self):
    #    return self
    @property
    def grid(self):
        return self.__grid__

    @property
    def n_nodes(self) -> int:
        return self.__nodes__.shape[0]

    def node(self, i: int):  # ->List:
        return self.__nodes__[i, :]

    @property
    def nodes(self):
        return self.__nodes__

    def n_inodes(self, i: int):  # ->int:
        return self.__inodes__.shape[1]

    def inode(self, i, j):  # ->List:
        return self.__inodes__[i, j]

    def iweight(self, i, j):  # ->float:
        return self.__iweights__[i, j]


class DiscretizedIIDProcess(DiscretizedProcess):
    def point(self, i: int) -> int:
        raise Exception("Not Implemented.")

    def weight(self, i: int) -> int:
        raise Exception("Not Implemented.")

    @property
    def grid(self):
        return EmptyGrid()

    @property
    def n_nodes(self) -> int:
        return 1  # purely conventional

    def node(self, i: int) -> Vector:
        return np.zeros(self.d)

    def n_inodes(self, i: int) -> int:
        return self.n

    def inode(self, i: int, j: int):  # vector
        return self.point(j)

    def iweight(self, i: int, j: int):  # scalar
        return self.weight(j)


@language_element
class MarkovChain(DiscretizedProcess, DiscreteProcess):

    signature = {"transitions": "Matrix", "values": "Matrix"}

    # transitions: 2d float array
    # values: 2d float array

    def __init__(self, transitions, values):

        self.transitions = np.array(transitions, dtype=np.float64)
        self.values = np.array(values, dtype=np.float64)
        self.d = self.values.shape[1]

    def discretize(self, to="mc"):

        if to == "gdp":
            nodes = self.values.copy()
            inodes = nodes[None, :, :].repeat(nodes.shape[0], axis=0)
            iweights = self.transitions
            return GDP(nodes, inodes, iweights)
            return
        elif to == "mc":
            return self
        else:
            raise Exception("Not implemented.")

    @property
    def grid(self):
        return UnstructuredGrid(self.values)

    @property
    def n_nodes(self):  # integer
        return self.values.shape[0]

    def node(self, i: int):  # vector
        return self.values[i, :]

    @property
    def nodes(self):
        return self.values

    def n_inodes(self, i: int):  # integer
        return self.transitions.shape[1]

    def inode(self, i: int, j: int):  # vector
        return self.values[j, :]

    def iweight(self, i: int, j: int):  # scalar
        return self.transitions[i, j]

    def simulate(self, N, T, i0=0, m0=None, stochastic=True):
        # m0 is basically ignored
        if stochastic:
            inds = simulate_markov_chain(self.values, self.transitions, N, T, i0=i0)
        else:
            inds = np.zeros((T, N), dtype=int) + i0
        return self.values[inds]


#%%


@language_element
class ProductProcess(Process):

    ### This class represents the product of processes

    def __init__(self, *l):
        self.processes = l
        self.d = sum([e.d for e in self.processes])

    def discretize(self, to=None, options={}):

        if isinstance(options, dict):
            kwargs = [options] * len(self.processes)
        else:
            assert len(options) == len(self.processes)
            kwargs = options

        if to is None:
            if any(
                [
                    isinstance(dp, MarkovChain) or isinstance(dp, ContinuousProcess)
                    for dp in self.processes
                ]
            ):
                to = "mc"
            elif all([isinstance(dp, IIDProcess) for dp in self.processes]):
                to = "iid"
            else:
                to = "gdp"

        if to == "iid":
            from dolo.numeric.distribution import product_iid

            fun = product_iid
        elif to == "mc":
            fun = product_mc
        elif to == "gdp":
            fun = product_gdp

        # discretize children
        discretized_processes = [
            e.discretize(to=to, **kwargs[i]) for i, e in enumerate(self.processes)
        ]

        return fun(discretized_processes)

    def simulate(self, N, T, m0=None, stochastic=True):

        if m0 is not None:
            raise Exception("Not implemented")
        sims = [
            p.simulate(N, T, m0=None, stochastic=stochastic) for p in self.processes
        ]
        return np.concatenate(sims, axis=2)


@language_element
def Product(*processes):
    return ProductProcess(*processes)


def product_mc(markov_chains: List[MarkovChain]) -> MarkovChain:

    M = [(m.values, m.transitions) for m in markov_chains]
    from dolo.numeric.discretization import tensor_markov

    [P, Q] = tensor_markov(*M)
    return MarkovChain(Q, P)


def product_gdp(gdps: List[GDP]) -> GDP:

    raise Exception("Not implemented")


@language_element
class AR1(ContinuousProcess):

    signature = {"ρ": "float", "σ": "float", "μ": "Optional[float]"}

    @greek_tolerance
    def __init__(self, ρ=None, σ=None, μ=None):
        self.ρ = ρ
        self.σ = σ
        self.μ = μ
        self.μ = μ if μ is not None else 0.0
        self.d = 1

    def discretize(self, N=3, to="mc", **kwargs):
        if to == "mc":
            return self.discretize_mc(N=N, **kwargs)
        elif to == "gdp":
            return self.discretize_gdp(N=N, **kwargs)

    def discretize_mc(self, N=3):

        rho = np.array([[self.ρ]])
        Sigma = np.array([[self.σ]])
        μ = np.array([self.μ])

        try:
            assert abs(np.eye(rho.shape[0]) * rho[0, 0] - rho).max() <= 1
        except:
            raise Exception(
                "When discretizing a Vector AR1 process, the autocorrelation coefficient must be as scalar. Found: {}".format(
                    rho_array
                )
            )

        from dolo.numeric.discretization import multidimensional_discretization

        [P, Q] = multidimensional_discretization(rho[0, 0], Sigma, N=N)

        P += μ[None, :]

        return MarkovChain(values=P, transitions=Q)

    def discretize_gdp(self, N=3):

        Σ = np.array([[self.σ]])
        ρ = self.ρ

        n_nodes = N
        n_std = 2.5
        n_integration_nodes = 5

        try:
            assert Σ.shape[0] == 1
        except:
            raise Exception("Not implemented.")

        try:
            assert ρ.shape[0] == ρ.shape[1] == 1
        except:
            raise Exception("Not implemented.")

        ρ = ρ[0, 0]
        σ = np.sqrt(Σ[0, 0])

        from dolo.numeric.discretization import gauss_hermite_nodes

        epsilons, weights = gauss_hermite_nodes([n_integration_nodes], Σ)

        min = -n_std * (σ / (np.sqrt(1 - ρ ** 2)))
        max = n_std * (σ / (np.sqrt(1 - ρ ** 2)))

        from .grids import CartesianGrid

        grid = CartesianGrid([min], [max], [n_nodes])

        nodes = np.linspace(min, max, n_nodes)[:, None]
        iweights = weights[None, :].repeat(n_nodes, axis=0)
        integration_nodes = np.zeros((n_nodes, n_integration_nodes))[:, :, None]
        for i in range(n_nodes):
            for j in range(n_integration_nodes):
                integration_nodes[i, j, :] = ρ * nodes[i, :] + epsilons[j]

        return GDP(nodes, integration_nodes, iweights, grid=grid)
        # return (nodes,integration_nodes,iweights)

    def simulate(self, N, T, m0=None, stochastic=True):

        Sigma = np.array([[self.σ]])
        mu = np.array([self.μ])
        rho = np.array([[self.ρ]])
        d = self.d

        if m0 is None:
            m0 = np.zeros(d)
        from numpy.random import multivariate_normal

        if stochastic:
            innov = multivariate_normal(mu, Sigma, N * T)
        else:
            innov = mu[None, len(mu)].repeat(T * N, axis=0)
        innov = innov.reshape((T, N, d))
        sim = np.zeros((T, N, d))
        sim[0, :, :] = m0[None, :]
        for t in range(1, sim.shape[0]):
            sim[t, :, :] = sim[t - 1, :, :] @ rho.T + innov[t, :, :]

        sim += μ[None, None, :]

        return sim

    def response(self, T, impulse):

        d = self.d
        rho = self.ρ
        μ = self.μ

        irf = np.zeros((T, d))
        irf[0, :] = impulse
        for t in range(1, irf.shape[0]):
            irf[t, :] = rho @ irf[t - 1, :]
        irf += μ[None, :]
        return irf


@language_element
class VAR1(ContinuousProcess):

    signature = {"ρ": "float", "Σ": "Matrix", "μ": "Optional[Vector]"}

    @greek_tolerance
    def __init__(self, ρ=None, Σ=None, μ=None):

        rho = ρ
        Sigma = Σ
        mu = μ

        self.Σ = np.atleast_2d(Sigma)
        d = self.Σ.shape[0]
        rho = np.array(rho)
        if rho.ndim == 0:
            self.ρ = np.eye(d) * rho
        elif rho.ndim == 1:
            self.ρ = np.diag(rho)
        else:
            self.ρ = rho
        if mu is None:
            self.μ = np.zeros(d)
        else:
            self.μ = np.array(mu, dtype=float)
        self.d = d

    def discretize(self, N=3, to="mc", **kwargs):
        if to == "mc":
            return self.discretize_mc(N=N, **kwargs)
        elif to == "gdp":
            return self.discretize_gdp(N=N, **kwargs)

    def discretize_mc(self, N=3):

        rho = self.ρ
        Sigma = self.Σ

        try:
            assert abs(np.eye(rho.shape[0]) * rho[0, 0] - rho).max() <= 1
        except:
            raise Exception(
                "When discretizing a Vector AR1 process, the autocorrelation coefficient must be as scalar. Found: {}".format(
                    rho_array
                )
            )

        from dolo.numeric.discretization import multidimensional_discretization

        [P, Q] = multidimensional_discretization(rho[0, 0], Sigma, N=N)

        P += self.μ[None, :]

        return MarkovChain(values=P, transitions=Q)

    def discretize_gdp(self, N=3):

        Σ = self.Σ
        ρ = self.ρ

        n_nodes = N
        n_std = 2.5
        n_integration_nodes = 5

        try:
            assert Σ.shape[0] == 1
        except:
            raise Exception("Not implemented.")

        try:
            assert ρ.shape[0] == ρ.shape[1] == 1
        except:
            raise Exception("Not implemented.")

        ρ = ρ[0, 0]
        σ = np.sqrt(Σ[0, 0])

        from dolo.numeric.discretization import gauss_hermite_nodes

        epsilons, weights = gauss_hermite_nodes([n_integration_nodes], Σ)

        min = -n_std * (σ / (np.sqrt(1 - ρ ** 2)))
        max = n_std * (σ / (np.sqrt(1 - ρ ** 2)))

        from .grids import CartesianGrid

        grid = CartesianGrid([min], [max], [n_nodes])

        nodes = np.linspace(min, max, n_nodes)[:, None]
        iweights = weights[None, :].repeat(n_nodes, axis=0)
        integration_nodes = np.zeros((n_nodes, n_integration_nodes))[:, :, None]
        for i in range(n_nodes):
            for j in range(n_integration_nodes):
                integration_nodes[i, j, :] = ρ * nodes[i, :] + epsilons[j]

        return GDP(nodes, integration_nodes, iweights, grid=grid)
        # return (nodes,integration_nodes,iweights)

    def simulate(self, N, T, m0=None, stochastic=True):

        d = self.d

        if m0 is None:
            m0 = np.zeros(d)
        from numpy.random import multivariate_normal

        Sigma = self.Σ
        mu = self.μ * 0
        if stochastic:
            innov = multivariate_normal(mu, Sigma, N * T)
        else:
            innov = mu[None, len(mu)].repeat(T * N, axis=0)
        innov = innov.reshape((T, N, d))
        rho = self.ρ
        sim = np.zeros((T, N, d))
        sim[0, :, :] = m0[None, :]
        for t in range(1, sim.shape[0]):
            sim[t, :, :] = sim[t - 1, :, :] @ rho.T + innov[t, :, :]

        sim += self.μ[None, None, :]

        return sim

    def response(self, T, impulse):
        d = self.d
        irf = np.zeros((T, d))
        irf[0, :] = impulse
        rho = self.ρ
        for t in range(1, irf.shape[0]):
            irf[t, :] = rho @ irf[t - 1, :]
        irf += self.μ[None, :]
        return irf


#### dummy class


@language_element
class Conditional(Process):

    signature = {"condition": "Any", "type": "Process", "arguments": "Function"}

    def __init__(self, condition=None, type=None, arguments=None):
        self.condition = condition
        self.type = type
        self.arguments = arguments



# I Go to Root.
def groot(path=""):
    "Changes current directory to the root of the project (looks for README.md)."

    def check_filelist(l):
        if ("README.md" in l) or ("README" in l):
            return True
        else:
            return False

    import os, copy

    cwd = os.getcwd()  # initial dir
    cwd0 = copy.copy(cwd)
    cwd_init = copy.copy(cwd)

    found = False
    sysroot = False

    while not found and not sysroot:
        found = check_filelist(os.listdir())
        if not found:
            os.chdir(os.path.join(cwd, os.pardir))
            cwd = os.getcwd()
            if cwd == cwd0:
                sysroot = True
            else:
                cwd0 = cwd
        else:
            os.chdir(os.path.join(cwd0, path))

    if sysroot:
        os.chdir(cwd_init)
        raise Exception("Root directory not found.")
